/*
  ex16.cu
  This is an exercise to demonstrate how to handle
  2-dimensional array in CUDA.
*/
#include <stdio.h>

__global__ void simple_kernel(float *input, float *output)
{
  int i = blockIdx.x * blockDim.x + threadIdx.x;

  output[i] = input[i];
}

int main(int argc, char *argv[])
{
  int i, j;
  /* Arrays in host memory */
  float **arr2D;
  float *arr1D;
  float *results;

  /* Arrays in device memory */
  float **cudaArr2D;
  float *cudaArr;
  float *cudaResults;

  int nrows = 8;
  int ncols = 4;

  /* The correct way to allocate 2D array in CPU */
  arr2D = (float **)malloc( nrows * sizeof(float *) );
  for( i = 0; i < nrows; i++ )
    arr2D[i] = (float *)malloc( ncols * sizeof(float) );

  for( i = 0; i < nrows; i++ )
  {
    for( j = 0 ; j < ncols; j++ )
    {
      arr2D[i][j] = (float) (i + j); 
    }
  }

  /* Wrong way to allocate 2D array in GPU, 
     execution of this code will not produce correct results. */
  cudaMalloc( (void **)cudaArr2D, nrows * sizeof(float *) );
  for( i = 0; i < nrows; i++ )
    cudaMalloc( (void **)&cudaArr2D[i], ncols * sizeof(float *) );

  /*******************************************************/
  /* Correct way for allocating and transferring memory */
  /*******************************************************/
#if 0
  /* First allocate a single pointer and copy in the contents */
  arr1D = (float *)malloc( nrows * ncols * sizeof(float) );
  results = (float *)malloc( nrows * ncols * sizeof(float) );

  for( i = 0; i < nrows; i++ )
    memcpy( &arr1D[i * ncols], arr2D[i], ncols * sizeof(float) );

  /* Allocate GPU memory */
  cudaMalloc( (void **)&cudaArr, nrows * ncols * sizeof(float) );
  cudaMalloc( (void **)&cudaResults, nrows * ncols * sizeof(float) );

  /* Transfer the data to the GPU */
  cudaMemcpy( cudaArr, arr1D, nrows * ncols * sizeof(float),
      cudaMemcpyHostToDevice );

  simple_kernel<<< nrows, ncols >>>(cudaArr, cudaResults);

  /* Memory transfer out */
  cudaMemcpy( results, cudaResults, nrows * ncols * sizeof(float),
      cudaMemcpyDeviceToHost );
#endif
  /* Print the results */
  printf( "Result matrix\n" );
  for( i = 0; i < nrows; i++ )
  {
    for( j = 0; j < ncols; j++ )
      printf( "%f ", results[i * ncols + j] );
    printf( "\n" );
  }

  /* Free the CPU 2D array */
  for( i = 0; i < nrows; i++ )
    free(arr2D[i]);
  free(arr2D);

  free(results);
  free(arr1D);
  cudaFree(cudaArr);
  cudaFree(cudaResults);

  return 0;
}
