/*
   ex33.cu. This is an exercise for looking at the concept of accumulating more data.
 */
#include <stdio.h>

void vector_add_cpu(const float *arr1, const float *arr2, float *results, int arrSize);

__global__ void vector_add_kernel(const float *arr1, const float *arr2, float *results)
{
  int index = blockIdx.x * blockDim.x + threadIdx.x;

  results[index] = arr1[index] + arr2[index];
}

int main(int argc, char *argv[])
{
  int i;
  int arrSize = 128;

  int batch = 1;

  if( argc > 1 ) batch = atoi(argv[1]);

  cudaEvent_t start, stop;
  float gpuTime = 0;

  float *arr1, *arr2;

  float *cudaArr1, *cudaArr2;
  float *cudaResults;
  float *resultsFromGPU;

  cudaEventCreate(&start);
  cudaEventCreate(&stop);

  /* Allocate CPU memory */
  arr1 = (float *)malloc( arrSize * batch * sizeof(float) );
  arr2 = (float *)malloc( arrSize * batch * sizeof(float) );
  resultsFromGPU = (float *)malloc( arrSize * batch * sizeof(float) );

  /* Allocate GPU memory */
  cudaMalloc( (void **)&cudaArr1, arrSize * batch * sizeof(float) );
  cudaMalloc( (void **)&cudaArr2, arrSize * batch * sizeof(float) );
  cudaMalloc( (void **)&cudaResults, arrSize * batch * sizeof(float) );

  /* Initialize vectors */
  for( i = 0; i < arrSize * batch; i++ )
  {
    arr1[i] = (float) (i * 3);
    arr2[i] = (float) ((i * 4) + 1);
  }

  /* CUDA implementation of vector addition */
  cudaEventRecord(start, 0);

  /* First copy out enough data */
  cudaMemcpy( cudaArr1, arr1, arrSize * batch * sizeof(float),
	cudaMemcpyHostToDevice );
  cudaMemcpy( cudaArr2, arr2, arrSize * batch * sizeof(float),
	cudaMemcpyHostToDevice );

  /* Choose the number of threads and blocks */
  dim3 threads(arrSize, 1);
  dim3 blocks(batch, 1);

  /* Kernel launch */
  vector_add_kernel<<< blocks, threads >>>(cudaArr1, cudaArr2, cudaResults);

  /* Copy the results back to CPU */
  cudaMemcpy( resultsFromGPU, cudaResults, arrSize * batch * sizeof(float), 
      cudaMemcpyDeviceToHost );

  cudaEventRecord(stop, 0);
  cudaEventSynchronize(stop);

  cudaEventElapsedTime( &gpuTime, start, stop );

  printf( "Execution time: %f (ms)\n", gpuTime );

  /* Free the memory */
  free(arr1);
  free(arr2);
  free(resultsFromGPU);

  cudaFree(cudaArr1);
  cudaFree(cudaArr2);
  cudaFree(cudaResults);

  return 0;
}

/* CPU implementation of vector addition */
void vector_add_cpu(const float *arr1, const float *arr2, float *resultsFromCPU, int arrSize)
{
  int i;
  for( i = 0; i < arrSize; i++ )
    resultsFromCPU[i] = arr1[i] + arr2[i];
}

